import asyncio
import re
from datetime import datetime
from unittest.mock import MagicMock

import freezegun
import pytest

from resnap.decorators import async_resnap, resnap
from resnap.exceptions import ResnapError
from resnap.helpers.context import add_metadata
from resnap.helpers.metadata import Metadata
from resnap.helpers.utils import hash_arguments
from tests.builders.metadata_builder import MetadataFailBuilder, MetadataSuccessBuilder


@pytest.fixture(autouse=True)
def mock_service(mocker) -> None:
    mock: MagicMock = mocker.patch("resnap.decorators.ResnapServiceFactory.get_service", return_value=MagicMock())
    return mock


@resnap
def func(magic_number: int = 40) -> int:
    if magic_number < 0:
        raise ValueError("Magic number should be positive")
    if magic_number == 0:
        raise ResnapError("Magic number should not be null", data={"key": "value"})
    return magic_number + 2


@resnap(consider_args=False)
def func_not_consider_args(magic_number: int = 40) -> int:
    if magic_number < 0:
        raise ValueError("Magic number should be positive")
    if magic_number == 0:
        raise ResnapError("Magic number should not be null", data={"key": "value"})
    return magic_number + 2


@resnap(enable_recovery=False)
def func_disable_recovery(magic_number: int = 40) -> int:
    if magic_number < 0:
        raise ValueError("Magic number should be positive")
    if magic_number == 0:
        raise ResnapError("Magic number should not be null", data={"key": "value"})
    return magic_number + 2


@resnap(output_folder="toto")
def func_output_folder(magic_number: int = 40) -> int:
    if magic_number < 0:
        raise ValueError("Magic number should be positive")
    if magic_number == 0:
        raise ResnapError("Magic number should not be null", data={"key": "value"})
    return magic_number + 2


@async_resnap
async def async_func(magic_number: int = 40) -> int:
    await asyncio.sleep(0.1)
    if magic_number < 0:
        raise ValueError("Magic number should be positive")
    if magic_number == 0:
        raise ResnapError("Magic number should not be null", data={"key": "value"})
    return magic_number + 2


@resnap(output_format="str")
def func_str(magic_number: int = 40) -> int:
    return magic_number + 2


@resnap
def func_str_with_custom_metadata(magic_number: int = 40) -> int:
    add_metadata("key", "value")
    return magic_number + 2


@async_resnap
async def async_func_str_with_custom_metadata(magic_number: int = 40) -> int:
    add_metadata("key", "value")
    await asyncio.sleep(0.1)
    return magic_number + 2


@async_resnap(output_format="str")
async def async_func_str(magic_number: int = 40) -> int:
    await asyncio.sleep(0.1)
    return magic_number + 2


def test_should_not_use_service_if_disabled_sync(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = False

    # When
    result = func()

    # Then
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_not_called()


def test_should_return_result_if_metadata_exists_sync(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = [
        MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 40}).build(),
        (
            MetadataSuccessBuilder.a_metadata()
            .with_event_time(datetime.fromisoformat("2021-01-02T00:00:00"))
            .with_arguments({"magic_number": 5})
            .build()
        ),
    ]
    mock_service.return_value.read_result.return_value = 42

    # When
    result = func()

    # Then
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_called_once()
    mock_service.return_value.get_success_metadata.assert_called_once()
    mock_service.return_value.read_result.assert_called_once()


@pytest.mark.parametrize(
    "metadata",
    [
        None,
        MetadataFailBuilder.a_metadata().build(),
        MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 10}).build(),
    ],
)
def test_should_return_result_sync(metadata: Metadata | None, mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = [metadata] if metadata else []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = func()

    # Then
    mock_service.return_value.read_result.assert_not_called()
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_called_once()
    mock_service.return_value.get_success_metadata.assert_called_once()
    mock_service.return_value.save_result.assert_called_with("func", 42, "", None)
    mock_service.return_value.save_success_metadata.assert_called_once_with(
        func_name="func",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": 40}),
        event_time=now_time,
        result_path="/path/to/result",
        result_type=int.__name__,
        extra_metadata={},
    )


def test_should_use_output_format_option_sync(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = func_str()

    # Then
    assert result == 42
    mock_service.return_value.save_result.assert_called_with("func_str", 42, "", "str")


def test_should_add_custom_metadata(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = func_str_with_custom_metadata()

    # Then
    assert result == 42
    mock_service.return_value.save_result.assert_called_with("func_str_with_custom_metadata", 42, "", None)
    mock_service.return_value.save_success_metadata.assert_called_once_with(
        func_name="func_str_with_custom_metadata",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": 40}),
        event_time=now_time,
        result_path="/path/to/result",
        result_type=int.__name__,
        extra_metadata={"key": "value"},
    )


def test_should_not_check_result_if_disabled(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = func_disable_recovery()

    # Then
    assert result == 42
    mock_service.return_value.read_result.assert_not_called()
    mock_service.return_value.save_result.assert_called_with("func_disable_recovery", 42, "", None)


def test_should_not_check_arguments_if_disabled(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.read_result.return_value = 30
    metadata = MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 30}).build()
    mock_service.return_value.get_success_metadata.return_value = [metadata]

    # When
    result = func_not_consider_args(12)

    # Then
    assert result == 30
    mock_service.return_value.read_result.assert_called_once_with(metadata)
    mock_service.return_value.save_result.assert_not_called()


@freezegun.freeze_time("2021-01-01")
@pytest.mark.parametrize(
    "magic_number, error_type, error_message, expected_data",
    [
        (0, ResnapError, "Magic number should not be null", {"key": "value"}),
        (-1, ValueError, "Magic number should be positive", {}),
    ],
)
def test_should_raise_error_sync(
    magic_number: int,
    error_type: Exception,
    error_message: str,
    expected_data: dict,
    mock_service: MagicMock,
) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []

    # When
    with pytest.raises(error_type, match=re.escape(error_message)):
        func(magic_number)

    # Then
    mock_service.return_value.save_failed_metadata.assert_called_once_with(
        func_name="func",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": magic_number}),
        event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
        error_message=error_message,
        data=expected_data,
        extra_metadata={},
    )


def test_should_use_output_folder_sync(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result/toto", now_time)

    # When
    result = func_output_folder()

    # Then
    assert result == 42
    mock_service.return_value.save_result.assert_called_with("func_output_folder", 42, "toto", None)
    mock_service.return_value.save_success_metadata.assert_called_once_with(
        func_name="func_output_folder",
        output_folder="toto",
        hashed_arguments=hash_arguments({"magic_number": 40}),
        event_time=now_time,
        result_path="/path/to/result/toto",
        result_type=int.__name__,
        extra_metadata={},
    )


@pytest.mark.asyncio
async def test_should_add_custom_metadata_async(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = await async_func_str_with_custom_metadata()

    # Then
    assert result == 42
    mock_service.return_value.save_result.assert_called_with("async_func_str_with_custom_metadata", 42, "", None)
    mock_service.return_value.save_success_metadata.assert_called_once_with(
        func_name="async_func_str_with_custom_metadata",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": 40}),
        event_time=now_time,
        result_path="/path/to/result",
        result_type=int.__name__,
        extra_metadata={"key": "value"},
    )


@pytest.mark.asyncio
async def test_should_not_use_service_if_disabled_async(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = False

    # When
    result = await async_func()

    # Then
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_not_called()


@pytest.mark.asyncio
async def test_should_return_result_if_metadata_exists_async(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = [
        MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 40}).build()
    ]
    mock_service.return_value.read_result.return_value = 42

    # When
    result = await async_func()

    # Then
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_called_once()
    mock_service.return_value.get_success_metadata.assert_called_once()
    mock_service.return_value.read_result.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "metadata",
    [
        None,
        MetadataFailBuilder.a_metadata().build(),
        MetadataSuccessBuilder.a_metadata().with_arguments({"magic_number": 10}).build(),
    ],
)
async def test_should_return_result_async(metadata: Metadata | None, mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = [metadata] if metadata else []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = await async_func()

    # Then
    mock_service.return_value.read_result.assert_not_called()
    assert result == 42
    mock_service.return_value.clear_old_saves.assert_called_once()
    mock_service.return_value.get_success_metadata.assert_called_once()
    mock_service.return_value.save_result.assert_called_with("async_func", 42, "", None)
    mock_service.return_value.save_success_metadata.assert_called_once_with(
        func_name="async_func",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": 40}),
        event_time=now_time,
        result_path="/path/to/result",
        result_type=int.__name__,
        extra_metadata={},
    )


@pytest.mark.asyncio
async def test_should_use_output_format_option_async(mock_service: MagicMock) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []
    now_time = datetime.now()
    mock_service.return_value.save_result.return_value = ("/path/to/result", now_time)

    # When
    result = await async_func_str()

    # Then
    assert result == 42
    mock_service.return_value.save_result.assert_called_with("async_func_str", 42, "", "str")


@freezegun.freeze_time("2021-01-01", real_asyncio=True)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "magic_number, error_type, error_message, expected_data",
    [
        (0, ResnapError, "Magic number should not be null", {"key": "value"}),
        (-1, ValueError, "Magic number should be positive", {}),
    ],
)
async def test_should_raise_error_async(
    magic_number: int,
    error_type: Exception,
    error_message: str,
    expected_data: dict,
    mock_service: MagicMock,
) -> None:
    # Given
    mock_service.return_value.is_enabled = True
    mock_service.return_value.get_success_metadata.return_value = []

    # When
    with pytest.raises(error_type, match=re.escape(error_message)):
        await async_func(magic_number)

    # Then
    mock_service.return_value.save_failed_metadata.assert_called_once_with(
        func_name="async_func",
        output_folder="",
        hashed_arguments=hash_arguments({"magic_number": magic_number}),
        event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
        error_message=error_message,
        data=expected_data,
        extra_metadata={},
    )



from threading import Thread

from resnap.helpers.singleton import SingletonABCMeta, SingletonMeta


class SingletonMetaClass(metaclass=SingletonMeta):
    pass


class SingletonABCMetaClass(metaclass=SingletonABCMeta):
    pass


def test_single_instance() -> None:
    instance1 = SingletonMetaClass()
    instance2 = SingletonMetaClass()
    assert instance1 is instance2, "SingletonMeta did not return the same instance"


def test_single_instance_abc() -> None:
    instance1 = SingletonABCMetaClass()
    instance2 = SingletonABCMetaClass()
    assert instance1 is instance2, "SingletonMeta did not return the same instance"


def test_thread_safety() -> None:
    instances = []

    def create_instance() -> None:
        instances.append(SingletonABCMetaClass())

    threads = [Thread(target=create_instance) for _ in range(10)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    first_instance = instances[0]
    for instance in instances:
        assert instance is first_instance, "SingletonMeta did not return the same instance across threads"



import inspect
from typing import Any, Callable


def get_function_signature(
    func: Callable,
    args: tuple,
    kwargs: dict,
    considered_attributes: list[str] | None = None,
) -> tuple[str, dict[str, Any]]:
    """
    Get the function name and arguments from the function signature.
    If the function is a method, the first argument is skipped and name is composed from the class and method name.

    Args:
        func (Callable): The function to get the signature from.
        args (tuple): The arguments passed to the function.
        kwargs (dict): The keyword arguments passed to the function.
        considered_attributes (list[str] | None): The list of class/instance attributes to consider when hashing
        the function arguments.
    Returns:
        tuple[str, dict[str, Any]]: The function name and the arguments.
    """
    considered_attributes = considered_attributes or []
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    func_name = ""
    arguments = {}
    for i, (name, value) in enumerate(bound_args.arguments.items()):
        if i == 0 and name in ("self", "cls"):
            arguments.update(get_attributes(value, considered_attributes))
            continue
        arguments[name] = value

    func_name = func.__qualname__
    return func_name, arguments


def _get_attributes(instance_attributes: dict[str, Any], considered_attributes: list[str]) -> dict[str, Any]:
    """
    Get the attributes of an instance.

    Args:
        instance (dict[str, Any]): The instance attributes.
        considered_attributes (list[str]): The list of class/instance attributes to consider when hashing the function
        arguments.
    Returns:
        dict[str, Any]: The attributes of the instance filtered.
    """
    return {
        attr: value for attr, value in instance_attributes
        if attr in considered_attributes
        and not callable(value)
        and not attr.startswith("__")
        and not isinstance(value, classmethod)
    }


def get_attributes(instance: Any, considered_attributes: list[str]) -> dict[str, Any]:
    """
    Get the attributes of an instance.

    Args:
        instance (Any): The instance to get the attributes from.
        considered_attributes (list[str]): The list of class/instance attributes to consider when hashing the function
        arguments.
    Returns:
        dict[str, Any]: The attributes of the instance.
    """
    class_attributes = _get_attributes(instance.__class__.__dict__.items(), considered_attributes)

    instance_attributes = _get_attributes(instance.__dict__.items(), considered_attributes)
    return {**class_attributes, **instance_attributes}



from typing import Any

import pytest
from pydantic import ValidationError

from resnap.helpers.config import Config, Services


class TestServices:
    def test_services(self) -> None:
        assert Services.LOCAL == "local"


class TestConfig:
    @pytest.mark.parametrize(
        "config",
        [
            pytest.param({"save_to": Services.LOCAL}, id="missing enabled"),
            pytest.param(
                {"save_to": Services.LOCAL, "enabled": 5}, id="wrong enabled type"
            ),
            pytest.param({"enabled": True, "save_to": "test"}, id="unknown save_to"),
            pytest.param(
                {"enabled": True, "save_to": Services.LOCAL, "output_base_path": 25},
                id="wrong output_base_path type",
            ),
            pytest.param(
                {"enabled": True, "save_to": Services.LOCAL, "secrets_file_name": 25},
                id="wrong secrets_file_name type",
            ),
            pytest.param(
                {
                    "enabled": True,
                    "save_to": Services.LOCAL,
                    "enable_remove_old_files": "test",
                },
                id="wrong enable_remove_old_files type",
            ),
            pytest.param(
                {
                    "enabled": True,
                    "save_to": Services.LOCAL,
                    "max_history_files_length": "test",
                },
                id="wrong max_history_files_length type",
            ),
            pytest.param(
                {
                    "enabled": True,
                    "save_to": Services.LOCAL,
                    "max_history_files_length": -5,
                },
                id="wrong max_history_files_length value",
            ),
            pytest.param(
                {
                    "enabled": True,
                    "save_to": Services.LOCAL,
                    "max_history_files_time_unit": 5,
                },
                id="wrong max_history_files_time_unit type",
            ),
            pytest.param(
                {
                    "enabled": True,
                    "save_to": Services.LOCAL,
                    "max_history_files_time_unit": "test",
                },
                id="wrong max_history_files_time_unit value",
            ),
        ],
    )
    def test_should_failed_with_wrong_config(self, config: dict[str, Any]) -> None:
        # When
        with pytest.raises(ValueError):
            Config(**config)

    def test_should_not_raise_if_not_enabled_and_save_to_is_ceph(self) -> None:
        # Given
        input_config = {"enabled": False}

        # When / Then
        Config(**input_config)

    def test_should_raise_if_not_secrets_file_name_and_save_to_is_boto(self) -> None:
        # When / Then
        with pytest.raises(ValidationError, match="secrets_file_name is required when save_to is s3"):
            Config(enabled=True, save_to=Services.S3)

    def test_should_build_and_check_config_successfully(self) -> None:
        # Given
        input_config = {
            "enabled": True,
            "save_to": Services.S3,
            "output_base_path": "",
            "secrets_file_name": "test",
            "enable_remove_old_files": True,
            "max_history_files_length": 5,
            "max_history_files_time_unit": "day",
        }
        # When
        config = Config(**input_config)

        # Then
        assert input_config == config.model_dump()



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from typing_extensions import Self

from .status import Status


@dataclass(frozen=True, kw_only=True)
class Metadata(ABC):
    status: Status
    event_time: datetime
    hashed_arguments: str
    extra_metadata: dict[str, Any] | None = None

    @classmethod
    def from_dict(cls, data: dict) -> MetadataSuccess | MetadataFail:
        if data["status"] == Status.SUCCESS.value:
            return MetadataSuccess.from_dict(data)
        return MetadataFail.from_dict(data)

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        raise NotImplementedError  # pragma: no cover


@dataclass(frozen=True, kw_only=True)
class MetadataSuccess(Metadata):
    result_path: str
    result_type: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(
            status=Status(data["status"]),
            event_time=datetime.fromisoformat(data["event_time"]),
            hashed_arguments=data["hashed_arguments"],
            result_path=data["result_path"],
            result_type=data["result_type"],
            extra_metadata=data.get("extra_metadata", {}),
        )

    def to_dict(self) -> dict[str, Any]:
        metadata = {
            "status": self.status.value,
            "event_time": self.event_time.isoformat(),
            "hashed_arguments": self.hashed_arguments,
            "result_path": self.result_path,
            "result_type": self.result_type,
        }
        if self.extra_metadata:
            metadata["extra_metadata"] = self.extra_metadata
        return metadata


@dataclass(frozen=True, kw_only=True)
class MetadataFail(Metadata):
    error_message: str
    data: dict

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(
            status=Status(data["status"]),
            event_time=datetime.fromisoformat(data["event_time"]),
            hashed_arguments=data["hashed_arguments"],
            error_message=data["error_message"],
            data=data.get("data", {}),
            extra_metadata=data.get("extra_metadata", {}),
        )

    def to_dict(self) -> dict[str, Any]:
        metadata = {
            "status": self.status.value,
            "event_time": self.event_time.isoformat(),
            "hashed_arguments": self.hashed_arguments,
            "error_message": self.error_message,
        }
        if self.data:
            metadata["data"] = self.data
        if self.extra_metadata:
            metadata["extra_metadata"] = self.extra_metadata
        return metadata



import gzip
import io
import re
from unittest.mock import MagicMock

import pandas as pd
import pytest

from resnap.boto.dataframes import (
    _SUPPORTED_COMPRESSIONS,
    AVAILABLE_COMPRESSIONS,
    CSVHandler,
    DataFrameHandler,
    ParquetHandler,
    get_dataframe_handler,
)


@pytest.mark.parametrize(
    "file_format, expected_handler",
    [
        ("csv", CSVHandler),
        ("parquet", ParquetHandler),
    ],
)
def test_should_return_dataframe_handler(
    file_format: str, expected_handler: DataFrameHandler
) -> None:
    # When
    handler = get_dataframe_handler(file_format)

    # Then
    assert handler == expected_handler


def test_should_raise_value_error_for_invalid_file_format() -> None:
    # Given
    invalid_file_format = "invalid_format"

    # When / Then
    with pytest.raises(
        ValueError,
        match="invalid_format is not supported. Valid ones are csv, parquet.",
    ):
        get_dataframe_handler(invalid_file_format)


@pytest.fixture
def mock_dataframe() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "col1": [1, 2],
            "col2": ["foo", "bar"],
        }
    )


class TestDataFrameHandler:
    @pytest.mark.parametrize(
        "compression, file_path",
        [
            (None, "test/dataframe.csv"),
            ("gzip", "test/dataframe.gz"),
            ("snappy", "test/dataframe.csv"),
        ],
    )
    def test_should_verify_compression_and_extension(
        self, compression: str | None, file_path: str
    ) -> None:
        # When
        DataFrameHandler.verify_compression_and_extension(compression, file_path)

        # Then
        assert True

    @pytest.mark.parametrize(
        "compression, file_path, expected_error_message",
        [
            pytest.param(
                "test",
                "test/dataframe.parquet",
                f"test not supported. It must be in this list: {', '.join(AVAILABLE_COMPRESSIONS)}",
                id="unsupported_compression",
            ),
            pytest.param(
                "gzip",
                "test/dataframe.csv",
                f"A gzip-compressed file must have an extension in this "
                f"list: {', '.join(AVAILABLE_COMPRESSIONS['gzip'])}",
                id="gzip_compression_invalid_extension",
            ),
        ],
    )
    def test_should_raise_value_error_for_invalid_compression_and_extension(
        self, compression: str, file_path: str, expected_error_message: str
    ) -> None:
        # When / Then
        with pytest.raises(ValueError, match=expected_error_message):
            DataFrameHandler.verify_compression_and_extension(compression, file_path)


class TestCSVHandler:
    def test_should_read_df(self, mock_dataframe: pd.DataFrame) -> None:
        # Given
        csv_content = b"col1,col2\n1,foo\n2,bar"
        fake_bytes_io = io.BytesIO(csv_content)

        # When
        result_df = CSVHandler.read_df(fake_bytes_io)

        # Then
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_read_df_nrows(self, mock_dataframe: pd.DataFrame) -> None:
        # Given
        lines = [
            b"col1,col2",
            b"1,foo",
            b"2,bar",
            b"3,baz",
        ]
        mock_bytes_obj = MagicMock()
        mock_bytes_obj.iter_lines.return_value = iter(lines)

        # When
        result_df = CSVHandler.read_df(mock_bytes_obj, nrows=2)

        # Then
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_read_df_nrows_gzip(self, mock_dataframe: pd.DataFrame) -> None:
        # Given
        csv_content = b"col1,col2\n1,foo\n2,bar\n3,baz\n4,qux\n"
        compressed_buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=compressed_buffer, mode="wb") as f:
            f.write(csv_content)
        compressed_bytes = compressed_buffer.getvalue()
        chunk_size = 10
        chunks = [
            compressed_bytes[i: i + chunk_size]
            for i in range(0, len(compressed_bytes), chunk_size)
        ]
        mock_bytes_obj = MagicMock()
        mock_bytes_obj.iter_chunks.return_value = iter(chunks)

        # When
        result_df = CSVHandler.read_df(mock_bytes_obj, compression="gzip", nrows=2)

        # Then
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_write_df_with_no_compression(
        self, mock_dataframe: pd.DataFrame
    ) -> None:
        # When
        buffer = CSVHandler.write_df(mock_dataframe, index=False)

        # Then
        assert isinstance(buffer, io.BytesIO)
        buffer.seek(0)
        result_df = pd.read_csv(buffer)
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_write_df_gzip_compression(
        self, mock_dataframe: pd.DataFrame
    ) -> None:
        # When
        buffer = CSVHandler.write_df(mock_dataframe, compression="gzip", index=False)

        # Then
        assert isinstance(buffer, io.BytesIO)
        buffer.seek(0)
        with gzip.GzipFile(fileobj=buffer, mode="rb") as f:
            decompressed = f.read()
        result_df = pd.read_csv(io.BytesIO(decompressed))
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_raise_valueerror_when_write_df_with_invalid_compression(
        self, mock_dataframe: pd.DataFrame
    ) -> None:
        # When / Then
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"test compression not supported, supported one are: {_SUPPORTED_COMPRESSIONS}"
            ),
        ):
            CSVHandler.write_df(mock_dataframe, compression="test")


class TestParquetHandler:
    def test_should_read_df(self, mock_dataframe: pd.DataFrame) -> None:
        # Given
        buffer = io.BytesIO()
        mock_dataframe.to_parquet(buffer, engine="pyarrow")
        buffer.seek(0)

        # When
        result_df = ParquetHandler.read_df(buffer)

        # Then
        pd.testing.assert_frame_equal(result_df, mock_dataframe)

    def test_should_write(self, mock_dataframe: pd.DataFrame) -> None:
        # When
        buffer = ParquetHandler.write_df(mock_dataframe)

        # Then
        assert isinstance(buffer, io.BytesIO)
        buffer.seek(0)
        result_df = pd.read_parquet(buffer, engine="pyarrow")
        pd.testing.assert_frame_equal(result_df, mock_dataframe)



import asyncio
from typing import Any

from container import JobContainer
from dependency_injector.wiring import Provide, inject


@inject
async def run_pipeline(conf_test: dict[str, Any], pipeline=Provide[JobContainer.pipeline]) -> None:
    await pipeline.run(
        first_method_argument=conf_test["first_method_argument"],
        second_method_argument=conf_test["second_method_argument"],
        third_method_argument=conf_test["third_method_argument"],
        basic_class_get_param_value_argument=conf_test["BasicClass_get_param_value_argument"],
        basic_class_generate_dataframe_argument=conf_test["BasicClass_generate_dataframe_argument"],
        a_function_argument=conf_test["a_function_argument"],
    )


def run_job(conf_test: dict[str, Any]) -> None:
    try:
        container: JobContainer = JobContainer()
        conf = {"test": {"test": True}}
        container.job_config.from_dict(conf)
        container.wire(
            modules=[
                __name__,
                "pipeline",
            ]
        )
        asyncio.run(run_pipeline(conf_test))
    except Exception as err:
        raise err



from unittest.mock import patch

import pytest
from packaging.version import parse as parse_version

import resnap
from resnap.version import version_short


def test_standard_version():
    v = parse_version(resnap.VERSION)
    assert str(v) == resnap.VERSION


def test_version_attribute_is_present():
    assert hasattr(resnap, "__version__")


def test_version_attribute_is_a_string():
    assert isinstance(resnap.__version__, str)


@pytest.mark.parametrize("version, expected", (("2.1", "2.1"), ("2.1.0", "2.1")))
def test_version_short(version, expected):
    with patch("resnap.version.VERSION", version):
        assert version_short() == expected



import json
import threading
from pathlib import Path
from typing import Any

import toml

from resnap import add_metadata, add_multiple_metadata, resnap

RESULT_PATH = Path(toml.load("pyproject.toml")["tool"]["resnap"]["output_base_path"])


def get_extra_metadata_in_resnap_file(key: str) -> dict[str, Any]:
    for f in RESULT_PATH.iterdir():
        if not f.is_file() or not f.name.endswith(".resnap"):
            continue
        with open(f, "r") as json_file:
            metadata = json.load(json_file)
            assert "extra_metadata" in metadata, (
                "extra_metadata key not found in resnap file"
            )
            if key in metadata["extra_metadata"]:
                return metadata["extra_metadata"]
    else:
        assert False, f"Resnap file not found for key {key}"


@resnap
def add_metadata_test(key: str = "test", value: Any = "value") -> None:
    add_metadata(key, value)


@resnap
def add_multiple_metadata_test() -> None:
    metadata = {
        "f": "foo",
        "b": "bar",
    }
    add_multiple_metadata(metadata)


def test_add_metadata() -> None:
    add_metadata_test()
    assert get_extra_metadata_in_resnap_file("test") == {"test": "value"}


def test_add_multiple_metadata() -> None:
    add_multiple_metadata_test()
    assert get_extra_metadata_in_resnap_file("f") == {"f": "foo", "b": "bar"}


def test_thread_metadata_isolation() -> None:
    threads = [
        threading.Thread(target=add_metadata_test, args=("x", "foo")),
        threading.Thread(target=add_metadata_test, args=("y", "bar")),
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    assert get_extra_metadata_in_resnap_file("x") == {"x": "foo"}
    assert get_extra_metadata_in_resnap_file("y") == {"y": "bar"}


def clear_project(path: Path = RESULT_PATH) -> None:
    if not RESULT_PATH.exists():
        return
    for file in path.iterdir():
        if file.is_file():
            file.unlink()
        else:
            clear_project(file)
    path.rmdir()


def run() -> None:
    print()
    print("**** START TESTS ADD METADATA ****")
    clear_project()

    print("Running test: execution with add_metadata")
    test_add_metadata()
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with add_multiple_metadata")
    test_add_multiple_metadata()
    print("Test passed")
    print("-" * 25)

    print("Running test: execution with add_metadata in threads")
    test_thread_metadata_isolation()
    print("Test passed")
    print("-" * 25)

    print("Cleaning files generated for tests")
    clear_project()
    print("-" * 25)

    print("**** END TESTS ADD METADATA ****")



import io
import json
import pickle
from datetime import datetime
from typing import Any

import pandas as pd

from ..boto import S3Client, S3Config
from ..helpers.config import Config
from ..helpers.constants import EXT, SEPARATOR
from ..helpers.metadata import Metadata
from ..helpers.status import Status
from ..helpers.utils import calculate_datetime_from_now, get_datetime_from_filename, load_file
from .service import ResnapService


class BotoResnapService(ResnapService):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self._client: S3Client = S3Client(
            S3Config(**load_file(config.secrets_file_name, key="resnap"))
        )

    @staticmethod
    def _format_path(path: str) -> str:
        return path if path.endswith(SEPARATOR) else f"{path}{SEPARATOR}"

    def clear_old_saves(self) -> None:
        if not self._client.object_exists(self._format_path(self.config.output_base_path)):
            return

        contents = self._client.list_objects(self.config.output_base_path, True)
        empty_folders: list[str] = self._clear_files(contents)
        if empty_folders:
            self._clear_folders(empty_folders)

    def _clear_files(self, files: list[str]) -> list[str]:
        limit_time: datetime = calculate_datetime_from_now(
            self.config.max_history_files_length, self.config.max_history_files_time_unit,
        )

        folders: set = set()
        to_delete: list = []
        for file in files:
            if file.endswith(SEPARATOR):
                folders.add(file)
                continue
            if EXT not in file:
                continue
            file_time = get_datetime_from_filename(file)
            if file_time < limit_time:
                to_delete.append(file)

        if to_delete:
            self._client.delete_objects(to_delete)
        return folders

    def _clear_folders(self, folders: set[str]) -> None:
        to_delete = []
        for folder in folders:
            contents = self._client.list_objects(folder)
            if not contents or (len(contents) == 1 and contents[0] == folder):
                to_delete.append(folder)
        if to_delete:
            self._client.delete_objects(to_delete)

    def _create_folder(self, path: str, folder_name: str) -> None:
        if path and not self._client.object_exists(self._format_path(path)):
            self._client.mkdir(path)
        if folder_name:
            output_path = SEPARATOR.join([path, folder_name])
            if not self._client.object_exists(self._format_path(output_path)):
                self._client.mkdir(output_path)

    def _read_metadata(self, metadata_path: str) -> Metadata:
        with self._get_buffer_for_read_file(metadata_path) as buffer:
            data = json.load(buffer)

        return Metadata.from_dict(data)

    def get_success_metadata(self, func_name: str, output_folder: str) -> list[Metadata]:
        files: list[str] = [
            file for file in self._client.list_objects(self._get_output_path(output_folder))
            if func_name in file and file.endswith(EXT)
        ]

        if not files:
            return []

        metadata: list[Metadata] = [self._read_metadata(f) for f in sorted(files, reverse=True)]
        return [m for m in metadata if m.status == Status.SUCCESS]

    def _read_parquet_to_dataframe(self, file_path: str) -> pd.DataFrame:
        return self._client.get_df_from_file(file_path, file_format="parquet")

    def _read_csv_to_dataframe(self, file_path: str) -> pd.DataFrame:
        return self._client.get_df_from_file(file_path, file_format="csv")

    def _get_buffer_for_read_file(self, file_path: str) -> io.BytesIO:
        buffer = io.BytesIO()
        self._client.download_file(buffer, file_path)
        buffer.seek(0)
        return buffer

    def _read_pickle(self, file_path: str) -> Any:
        with self._get_buffer_for_read_file(file_path) as buffer:
            return pickle.loads(buffer.read())

    def _read_text(self, file_path: str) -> str:
        with self._get_buffer_for_read_file(file_path) as buffer:
            return buffer.read().decode()

    def _read_json(self, file_path: str) -> Any:
        with self._get_buffer_for_read_file(file_path) as buffer:
            return json.load(buffer)

    def _save_dataframe_to_csv(self, result: pd.DataFrame, result_path: str) -> None:
        self._client.push_df_to_file(result, result_path, file_format="csv")

    def _save_dataframe_to_parquet(self, result: pd.DataFrame, result_path: str) -> None:
        self._client.push_df_to_file(result, result_path, compression="gzip", file_format="parquet")

    def _save_to_pickle(self, result: Any, result_path: str) -> None:
        with io.BytesIO() as buffer:
            pickle.dump(result, buffer)
            self._client.upload_file(buffer, result_path)

    def _save_to_text(self, result: Any, result_path: str) -> None:
        with io.BytesIO() as buffer:
            buffer.write(str(result).encode())
            self._client.upload_file(buffer, result_path)

    def _save_to_json(self, result: Any, result_path: str) -> None:
        with io.BytesIO() as buffer:
            buffer.write(json.dumps(result, indent=4).encode())
            self._client.upload_file(buffer, result_path)

    def _write_metadata(self, metadata_path: str, metadata: Metadata) -> None:
        with io.BytesIO() as buffer:
            buffer.write(json.dumps(metadata.to_dict(), indent=4).encode())
            self._client.upload_file(buffer, metadata_path)



from .helpers.config import Config, Services
from .services.local_service import LocalResnapService
from .services.service import ResnapService
from .settings import get_config_data

_resnap_config: Config = get_config_data()
_service: ResnapService | None = None


def set_resnap_service(service: ResnapService) -> None:
    """
    Set the resnap service.

    Args:
        service (ResnapService): Resnap service.
    """
    if not isinstance(service, ResnapService):
        raise TypeError(f"Expected ResnapService, got {type(service)}")
    global _service
    _service = service


class ResnapServiceFactory:
    @classmethod
    def get_service(cls) -> ResnapService:
        """
        Get resnap service based on the configuration.

        Returns:
            ResnapService: Resnap service.
        """
        global _service
        if _service is not None:
            return _service
        if _resnap_config.save_to == Services.LOCAL:
            _service = LocalResnapService(_resnap_config)
        elif _resnap_config.save_to == Services.S3:
            from .services.boto_service import BotoResnapService
            _service = BotoResnapService(_resnap_config)
        else:
            raise NotImplementedError(f"Resnap service {_resnap_config.save_to} is not implemented")
        return _service



import json
import pickle
from datetime import datetime
from unittest.mock import ANY, MagicMock

import pandas as pd
import pytest

from resnap.helpers.metadata import Metadata, MetadataSuccess
from resnap.helpers.status import Status
from resnap.helpers.utils import hash_arguments
from resnap.services.boto_service import BotoResnapService
from tests.builders.config_builder import ConfigBuilder

s3_secrets: dict = {
    "endpoint_url": "http://s3-server",
    "access_key": "access_key",
    "secret_key": "secret_key",
    "bucket_name": "bucket",
}


@pytest.fixture(autouse=True)
def mock_load_file(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch(
        "resnap.services.boto_service.load_file", return_value=s3_secrets
    )
    return mock


@pytest.fixture
def mock_s3_client_object_exists(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.object_exists")
    return mock


@pytest.fixture
def mock_s3_client_list_objects(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.list_objects")
    return mock


@pytest.fixture
def mock_s3_client_delete_objects(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.delete_objects")
    return mock


@pytest.fixture
def mock_s3_client_download_file(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.download_file")
    return mock


@pytest.fixture
def mock_s3_client_upload_file(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.upload_file")
    return mock


@pytest.fixture
def mock_s3_client_push_df_to_file(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.push_df_to_file")
    return mock


@pytest.fixture
def mock_s3_client_get_df_from_file(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.get_df_from_file")
    return mock


@pytest.fixture
def mock_s3_client_mkdir(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.S3Client.mkdir")
    return mock


@pytest.fixture
def mock_read_metadata(mocker) -> MagicMock:
    mock: MagicMock = mocker.patch("resnap.services.boto_service.BotoResnapService._read_metadata")
    return mock


class TestBotoService:
    @pytest.mark.parametrize(
        "path, expected",
        [
            ("folder/", "folder/"),
            ("folder", "folder/"),
        ],
    )
    def test_should_format_path(self, path: str, expected: str) -> None:
        # When
        result = BotoResnapService._format_path(path)

        # Then
        assert result == expected

    @pytest.mark.parametrize(
        "file_path, is_deleted",
        [
            ("test_2021-01-01T00-00-00.resnap", True),
            ("test_2021-01-01T00-00-00.resnap.pkl", True),
            ("toto/toto_2021-01-01T00-00-00.csv", False),
            (f"test_{datetime.now().isoformat().replace(':', '-')}.resnap", False),
        ],
    )
    def test_should_clear_old_saves_files(
        self,
        file_path: str,
        is_deleted: bool,
        mock_s3_client_object_exists: MagicMock,
        mock_s3_client_list_objects: MagicMock,
        mock_s3_client_delete_objects: MagicMock,
    ) -> None:
        # Given
        mock_s3_client_object_exists.return_value = True
        mock_s3_client_list_objects.return_value = [file_path]
        service = BotoResnapService(ConfigBuilder.a_config().build())

        # When
        service.clear_old_saves()

        # Then
        mock_s3_client_list_objects.assert_called_once()
        assert mock_s3_client_delete_objects.call_count == int(is_deleted)

    @pytest.mark.parametrize(
        "file_path, is_empty, is_deleted",
        [
            ("toto/", False, False),
            ("toto/", True, True),
        ],
    )
    def test_should_clear_old_saves_folders(
        self,
        file_path: str,
        is_empty: bool,
        is_deleted: bool,
        mock_s3_client_object_exists: MagicMock,
        mock_s3_client_list_objects: MagicMock,
        mock_s3_client_delete_objects: MagicMock,
    ) -> None:
        # Given
        side_effects = [[file_path]]
        if is_empty:
            side_effects.append([])
        else:
            side_effects.append(["test.csv"])
        mock_s3_client_list_objects.side_effect = side_effects
        mock_s3_client_object_exists.return_value = True
        service = BotoResnapService(ConfigBuilder.a_config().build())

        # When
        service.clear_old_saves()

        # Then
        assert mock_s3_client_list_objects.call_count == 2
        assert mock_s3_client_delete_objects.call_count == int(is_deleted)

    def test_should_not_clear_old_saves_if_folder_does_not_exist(
        self,
        mock_s3_client_object_exists: MagicMock,
        mock_s3_client_list_objects: MagicMock,
    ) -> None:
        # Given
        mock_s3_client_object_exists.return_value = False
        service = BotoResnapService(ConfigBuilder.a_config().build())

        # When
        service.clear_old_saves()

        # Then
        mock_s3_client_list_objects.assert_not_called()

    def test_should_read_metadata(self, mock_s3_client_download_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        expected_metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            result_path="test_2021-01-01T00-00-00.resnap.pkl",
            result_type="str",
            hashed_arguments=hash_arguments({"test": "toto"}),
            extra_metadata={},
        )
        mock_s3_client_download_file.side_effect = lambda buffer, path: buffer.write(
            json.dumps(expected_metadata.to_dict()).encode()
        )

        # When
        result = service._read_metadata("tests/data/metadata/test-metadata_2021-01-01T00-00-00.resnap")

        # Then
        assert isinstance(result, Metadata)
        assert result == expected_metadata

    def test_should_return_multiple_metadata(
        self,
        mock_s3_client_list_objects: MagicMock,
        mock_read_metadata: MagicMock,
    ) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        mock_s3_client_list_objects.return_value = [
            "test-metadata_2021-01-01T00-00-00.resnap",
            "test-metadata_2024-01-01T00-00-00.resnap",
            "toto",
            "toto_2021-01-02T00-00-00.resnap",
            "test.csv",
        ]
        expected_metadata = [
            MetadataSuccess(
                status=Status.SUCCESS,
                event_time=datetime.fromisoformat("2024-01-01T00:00:00"),
                result_path="test_2024-01-01T00-00-00.resnap.pkl",
                result_type="str",
                hashed_arguments=hash_arguments({}),
            ),
            MetadataSuccess(
                status=Status.SUCCESS,
                event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
                result_path="test_2021-01-01T00-00-00.resnap.pkl",
                result_type="str",
                hashed_arguments=hash_arguments({}),
            ),
        ]
        mock_read_metadata.side_effect = expected_metadata

        # When
        result = service.get_success_metadata("test", "")

        # Then
        assert result == expected_metadata
        assert len(result) == 2

    def test_should_return_none_if_not_metadata(self, mock_s3_client_list_objects: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        mock_s3_client_list_objects.return_value = []

        # When
        result = service.get_success_metadata("test", "")

        # Then
        assert len(result) == 0

    def test_should_write_metadata(self, mock_s3_client_upload_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        metadata = MetadataSuccess(
            status=Status.SUCCESS,
            event_time=datetime.fromisoformat("2021-01-01T00:00:00"),
            result_path="test_2021-01-01T00-00-00.resnap.pkl",
            result_type="str",
            hashed_arguments=hash_arguments({"test": "toto"}),
        )
        mock_s3_client_upload_file.side_effect = lambda buffer, path: buffer.write(
            json.dumps(metadata.to_dict()).encode()
        )

        # When
        service._write_metadata("test_2021-01-01T00-00-00.resnap", metadata)

        # Then
        mock_s3_client_upload_file.assert_called_once()
        args, _ = mock_s3_client_upload_file.call_args
        assert args[1] == "test_2021-01-01T00-00-00.resnap"

    def test_should_save_dataframe_to_parquet(self, mock_s3_client_push_df_to_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        result = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        result_path = "test.parquet"

        # When
        service._save_dataframe_to_parquet(result, result_path)

        # Then
        mock_s3_client_push_df_to_file.assert_called_once_with(
            result, "test.parquet", compression="gzip", file_format="parquet"
        )

    def test_should_save_dataframe_to_csv(self, mock_s3_client_push_df_to_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        result = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        result_path = "test.csv"

        # When
        service._save_dataframe_to_csv(result, result_path)

        # Then
        mock_s3_client_push_df_to_file.assert_called_once_with(result, "test.csv", file_format="csv")

    def test_should_save_to_pickle(self, mock_s3_client_upload_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        result = {"key": "value"}
        result_path = "test.pkl"

        # When
        service._save_to_pickle(result, result_path)

        # Then
        mock_s3_client_upload_file.assert_called_once()
        args, _ = mock_s3_client_upload_file.call_args
        assert args[1] == "test.pkl"

    def test_should_save_to_text(self, mock_s3_client_upload_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        result = "toto"
        result_path = "test.txt"

        # When
        service._save_to_text(result, result_path)

        # Then
        mock_s3_client_upload_file.assert_called_once()
        args, _ = mock_s3_client_upload_file.call_args
        assert args[1] == "test.txt"

    def test_should_save_to_json(self, mock_s3_client_upload_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        result = {"key": "value"}
        result_path = "test.json"

        # When
        service._save_to_json(result, result_path)

        # Then
        mock_s3_client_upload_file.assert_called_once()
        args, _ = mock_s3_client_upload_file.call_args
        assert args[1] == "test.json"

    def test_should_read_parquet_to_dataframe(self, mock_s3_client_get_df_from_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        file_path = "test.parquet"
        expected_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        mock_s3_client_get_df_from_file.return_value = expected_df

        # When
        result = service._read_parquet_to_dataframe(file_path)

        # Then
        pd.testing.assert_frame_equal(result, expected_df)
        mock_s3_client_get_df_from_file.assert_called_once_with("test.parquet", file_format="parquet")

    def test_should_read_csv_to_dataframe(self, mock_s3_client_get_df_from_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        file_path = "test.csv"
        expected_df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
        mock_s3_client_get_df_from_file.return_value = expected_df

        # When
        result = service._read_csv_to_dataframe(file_path)

        # Then
        pd.testing.assert_frame_equal(result, expected_df)
        mock_s3_client_get_df_from_file.assert_called_once_with("test.csv", file_format="csv")

    def test_should_read_pickle(self, mock_s3_client_download_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        file_path = "test.pkl"
        expected_data = {"key": "value"}
        mock_s3_client_download_file.side_effect = lambda buffer, path: buffer.write(pickle.dumps(expected_data))

        # When
        result = service._read_pickle(file_path)

        # Then
        assert result == expected_data
        mock_s3_client_download_file.assert_called_once_with(ANY, str(file_path))

    def test_should_read_text(self, mock_s3_client_download_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        file_path = "test.txt"
        expected_data = "toto"
        mock_s3_client_download_file.side_effect = lambda buffer, path: buffer.write(expected_data.encode())

        # When
        result = service._read_text(file_path)

        # Then
        assert result == expected_data
        mock_s3_client_download_file.assert_called_once_with(ANY, str(file_path))

    def test_should_read_json(self, mock_s3_client_download_file: MagicMock) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        file_path = "test.json"
        expected_data = {"key": "value"}
        mock_s3_client_download_file.side_effect = (
            lambda buffer, path: buffer.write(json.dumps(expected_data, indent=4).encode())
        )

        # When
        result = service._read_json(file_path)

        # Then
        assert result == expected_data
        mock_s3_client_download_file.assert_called_once_with(ANY, str(file_path))

    @pytest.mark.parametrize(
        "path, is_exists_path, folder_name, is_exists_folder, exist_count, mkdir_count",
        [
            ("toto", False, "folder", False, 2, 2),
            ("toto", True, "folder", False, 2, 1),
            ("toto", True, "folder", True, 2, 0),
            ("toto", False, "", False, 1, 1),
            ("toto", True, "", False, 1, 0),
            ("", False, "folder", False, 1, 1),
            ("", False, "folder", True, 1, 0),
        ],
    )
    def test_should_create_folder(
        self,
        path: str,
        is_exists_path: bool,
        folder_name: str,
        is_exists_folder: bool,
        exist_count: int,
        mkdir_count: int,
        mock_s3_client_mkdir: MagicMock,
        mock_s3_client_object_exists: MagicMock,
    ) -> None:
        # Given
        service = BotoResnapService(ConfigBuilder.a_config().build())
        side_effects = []
        if path:
            side_effects.append(is_exists_path)
        if folder_name:
            side_effects.append(is_exists_folder)
        mock_s3_client_object_exists.side_effect = side_effects

        # When
        service._create_folder(path, folder_name)

        # Then
        assert mock_s3_client_object_exists.call_count == exist_count
        assert mock_s3_client_mkdir.call_count == mkdir_count



from typing import Any

import pytest
from pydantic import ValidationError

from resnap.boto.config import S3Config


class TestS3Config:
    @pytest.mark.parametrize(
        "config",
        [
            pytest.param(
                {"access_key": "key", "secret_key": "secret"}, id="missing bucket_name"
            ),
            pytest.param(
                {"access_key": "key", "bucket_name": "bucket"}, id="missing secret_key"
            ),
            pytest.param(
                {"secret_key": "secret", "bucket_name": "bucket"},
                id="missing access_key",
            ),
            pytest.param(
                {
                    "access_key": "key",
                    "secret_key": "secret",
                    "bucket_name": "bucket",
                    "force_path_style": "invalid",
                },
                id="invalid force_path_style type",
            ),
            pytest.param(
                {
                    "access_key": "key",
                    "secret_key": "secret",
                    "bucket_name": "bucket",
                    "signature_version": 123,
                },
                id="invalid signature_version type",
            ),
        ],
    )
    def test_s3_config_should_raise_validation_error(self, config: dict[str, Any]) -> None:
        # When / Then
        with pytest.raises(ValidationError):
            S3Config(**config)

    def test_s3_config_defaults(self) -> None:
        # Given
        config = {
            "access_key": "key",
            "secret_key": "secret",
            "bucket_name": "bucket",
        }

        # When
        s3_config = S3Config(**config)

        # Then
        assert s3_config.region_name is None
        assert s3_config.endpoint_url is None
        assert s3_config.force_path_style is True
        assert s3_config.signature_version == "s3v4"



from abc import ABC
from datetime import datetime
from typing import Any

from typing_extensions import Self

from resnap.helpers.metadata import MetadataFail, MetadataSuccess
from resnap.helpers.status import Status
from resnap.helpers.utils import hash_arguments


class MetadataBuilder(ABC):
    _event_time: datetime = datetime.fromisoformat("2021-01-01T00:00:00")
    _hashed_arguments: str = ""
    _extra_metadata: dict[str, Any] = {}

    @classmethod
    def a_metadata(cls) -> Self:
        return cls()

    def with_event_time(self, event_time: datetime) -> Self:
        self._event_time = event_time
        return self

    def with_arguments(self, arguments: dict[str, Any]) -> Self:
        self._hashed_arguments = hash_arguments(arguments)
        return self

    def with_extra_metadata(self, extra_metadata: dict[str, Any]) -> Self:
        self._extra_metadata = extra_metadata
        return self


class MetadataSuccessBuilder(MetadataBuilder):
    _status: Status = Status.SUCCESS
    _result_path: str = "/path/to/result"
    _result_type: str = "str"

    def with_result_path(self, result_path: str) -> Self:
        self._result_path = result_path
        return self

    def with_result_type(self, result_type: str) -> Self:
        self._result_type = result_type
        return self

    def build(self) -> MetadataSuccess:
        return MetadataSuccess(
            status=self._status,
            event_time=self._event_time,
            hashed_arguments=self._hashed_arguments,
            result_path=self._result_path,
            result_type=self._result_type,
            extra_metadata=self._extra_metadata,
        )


class MetadataFailBuilder(MetadataBuilder):
    _status: Status = Status.FAIL
    _error_message: str = ""
    _data: dict = {}

    def with_error_message(self, error_message: str) -> Self:
        self._error_message = error_message
        return self

    def with_data(self, data: dict) -> Self:
        self._data = data
        return self

    def build(self) -> MetadataFail:
        return MetadataFail(
            status=self._status,
            event_time=self._event_time,
            hashed_arguments=self._hashed_arguments,
            error_message=self._error_message,
            data=self._data,
            extra_metadata=self._extra_metadata,
        )
